"""Implementation of clock provider functionality."""

from __future__ import annotations

import uuid
from typing import TYPE_CHECKING

from sdc11073.provider.operations import ExecuteResult

from .nomenclature import NomenclatureCodes
from .providerbase import OperationClassGetter, ProviderRole

if TYPE_CHECKING:
    from sdc11073.mdib.descriptorcontainers import AbstractOperationDescriptorProtocol
    from sdc11073.mdib.mdibprotocol import ProviderMdibProtocol
    from sdc11073.provider.operations import ExecuteParameters, OperationDefinitionBase
    from sdc11073.provider.sco import AbstractScoOperationsRegistry


class GenericSDCClockProvider(ProviderRole):
    """Handles operations for setting ntp server and time zone.

    This provider handles SetString operations with codes
    "MDC_OP_SET_TIME_SYNC_REF_SRC" and "MDC_ACT_SET_TIME_ZONE".
    Nothing is added to the mdib. If the mdib does not contain these operations, the functionality is not available.
    """

    def __init__(self, mdib: ProviderMdibProtocol, log_prefix: str):
        super().__init__(mdib, log_prefix)
        self._set_ntp_operations = []
        self._set_tz_operations = []
        pm_types = self._mdib.data_model.pm_types

        self.MDC_OP_SET_TIME_SYNC_REF_SRC = pm_types.CodedValue(NomenclatureCodes.MDC_OP_SET_TIME_SYNC_REF_SRC)
        self.MDC_ACT_SET_TIME_ZONE = pm_types.CodedValue(NomenclatureCodes.MDC_ACT_SET_TIME_ZONE)

    def init_operations(self, sco: AbstractScoOperationsRegistry):
        """Create a ClockDescriptor and ClockState in mdib if they do not exist in mdib."""
        super().init_operations(sco)
        pm_types = self._mdib.data_model.pm_types
        pm_names = self._mdib.data_model.pm_names
        clock_entities = self._mdib.entities.by_node_type(pm_names.ClockDescriptor)
        if len(clock_entities) == 0:
            mds_entities = self._mdib.entities.by_node_type(pm_names.MdsDescriptor)
            if len(mds_entities) == 0:
                self._logger.info('empty mdib, cannot create a clock descriptor')
                return
            # create a clock descriptor for the first mds
            my_mds_entity = mds_entities[0]
            clock_descr_handle = 'clock_' + uuid.uuid4().hex
            self._logger.debug('creating a clock descriptor, handle=%s', clock_descr_handle)
            model = self._mdib.data_model
            clock_entity = self._mdib.entities.new_entity(
                model.pm_names.ClockDescriptor,
                handle=clock_descr_handle,
                parent_handle=my_mds_entity.handle,
            )
            clock_entity.descriptor.SafetyClassification = pm_types.SafetyClassification.INF
            clock_entity.descriptor.Type = pm_types.CodedValue('123')
            with self._mdib.descriptor_transaction() as mgr:
                mgr.write_entity(clock_entity)

    def make_operation_instance(
        self,
        operation_descriptor_container: AbstractOperationDescriptorProtocol,
        operation_cls_getter: OperationClassGetter,
    ) -> OperationDefinitionBase | None:
        """Create operation handlers.

        Handle codes MDC_OP_SET_TIME_SYNC_REF_SRC, MDC_ACT_SET_TIME_ZONE.
        """
        if operation_descriptor_container.coding == self.MDC_OP_SET_TIME_SYNC_REF_SRC.coding:
            self._logger.debug(
                'instantiating "set ntp server" operation from existing descriptor handle=%s',
                operation_descriptor_container.Handle,
            )
            set_ntp_operation = self._mk_operation_from_operation_descriptor(
                operation_descriptor_container,
                operation_cls_getter,
                operation_handler=self._set_ntp_string,
            )
            self._set_ntp_operations.append(set_ntp_operation)
            return set_ntp_operation
        if operation_descriptor_container.coding == self.MDC_ACT_SET_TIME_ZONE.coding:
            self._logger.debug(
                'instantiating "set time zone" operation from existing descriptor handle=%s',
                operation_descriptor_container.Handle,
            )
            set_tz_operation = self._mk_operation_from_operation_descriptor(
                operation_descriptor_container,
                operation_cls_getter,
                operation_handler=self._set_tz_string,
            )
            self._set_tz_operations.append(set_tz_operation)
            return set_tz_operation
        return None

    def _set_ntp_string(self, params: ExecuteParameters) -> ExecuteResult:
        """Set the ReferenceSource value of clock state (ExecuteHandler)."""
        value = params.operation_request.argument
        pm_names = self._mdib.data_model.pm_names
        self._logger.info(
            'set value %s from %s to %s',
            params.operation_instance.operation_target_handle,
            params.operation_instance.current_value,
            value,
        )

        op_target_entity = self._mdib.entities.by_handle(params.operation_instance.operation_target_handle)

        # look for clock entities that are a direct child of this mds
        mds_handle = op_target_entity.descriptor.source_mds or op_target_entity.handle
        clock_entities = self._mdib.entities.by_node_type(pm_names.ClockDescriptor)
        clock_entities = [c for c in clock_entities if c.parent_handle == mds_handle]

        if len(clock_entities) == 0:
            self._logger.warning('_set_ntp_string: no clock entity found')
            return ExecuteResult(
                params.operation_instance.operation_target_handle,
                self._mdib.data_model.msg_types.InvocationState.FAILED,
            )

        clock_entities[0].state.ReferenceSource = [value]
        with self._mdib.component_state_transaction() as mgr:
            mgr.write_entity(clock_entities[0])
        return ExecuteResult(
            params.operation_instance.operation_target_handle,
            self._mdib.data_model.msg_types.InvocationState.FINISHED,
        )

    def _set_tz_string(self, params: ExecuteParameters) -> ExecuteResult:
        """Set the TimeZone value of clock state (ExecuteHandler)."""
        value = params.operation_request.argument
        pm_names = self._mdib.data_model.pm_names
        self._logger.info(
            'set value %s from %s to %s',
            params.operation_instance.operation_target_handle,
            params.operation_instance.current_value,
            value,
        )

        op_target_entity = self._mdib.entities.by_handle(params.operation_instance.operation_target_handle)

        # look for clock entities that are a direct child of this mds
        mds_handle = op_target_entity.descriptor.source_mds or op_target_entity.handle
        clock_entities = self._mdib.entities.by_node_type(pm_names.ClockDescriptor)
        clock_entities = [c for c in clock_entities if c.parent_handle == mds_handle]

        if len(clock_entities) == 0:
            self._logger.warning('_set_ntp_string: no clock entity found')
            return ExecuteResult(
                params.operation_instance.operation_target_handle,
                self._mdib.data_model.msg_types.InvocationState.FAILED,
            )

        clock_entities[0].state.TimeZone = value
        with self._mdib.component_state_transaction() as mgr:
            mgr.write_entity(clock_entities[0])
        return ExecuteResult(
            params.operation_instance.operation_target_handle,
            self._mdib.data_model.msg_types.InvocationState.FINISHED,
        )



"""The module implements the different kinds of BICEPS operations.

These operations are the instances inside a provider that perform an operation by changing the mdib content.
"""
from __future__ import annotations

import inspect
import sys
import time
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Protocol

from sdc11073 import loghelper
from sdc11073 import observableproperties as properties
from sdc11073.exceptions import ApiUsageError
from sdc11073.xml_types import pm_qnames as pm

if TYPE_CHECKING:
    from lxml import etree

    from sdc11073.mdib.descriptorcontainers import AbstractDescriptorProtocol
    from sdc11073.mdib.providermdib import ProviderMdib
    from sdc11073.pysoap.soapenvelope import ReceivedSoapMessage
    from sdc11073.xml_types.msg_types import AbstractSet, InvocationState
    from sdc11073.xml_types.pm_types import CodedValue, OperatingMode


class OperationDefinitionProtocol(Protocol):
    """Interface that ExecuteHandlers need."""

    handle: str
    operation_target_handle: str
    current_value: Any
    last_called_time: float | None
    descriptor_container: AbstractDescriptorProtocol


@dataclass
class ExecuteParameters:
    """The argument of the ExecuteHandler call."""

    operation_instance: OperationDefinitionProtocol
    operation_request: AbstractSet
    soap_message: ReceivedSoapMessage


@dataclass
class ExecuteResult:
    """The return value of the ExecuteHandler call."""

    operation_target_handle: str
    invocation_state: InvocationState  # = InvocationState.FINISHED # only return a final state, not WAIT or STARTED


# ExecuteHandler also get the full soap message as parameter, because the soap header might contain
# relevant information, e.g. safety requirements.
ExecuteHandler = Callable[[ExecuteParameters], ExecuteResult]
TimeoutHandler = Callable[[OperationDefinitionProtocol], None]


class OperationDefinitionBase:
    """Base class of all provided operations.

    An operation is a point for remote control over the network.
    """

    current_value: Any = properties.ObservableProperty(fire_only_on_changed_value=False)
    current_request = properties.ObservableProperty(fire_only_on_changed_value=False)
    current_argument = properties.ObservableProperty(fire_only_on_changed_value=False)
    on_timeout = properties.ObservableProperty(fire_only_on_changed_value=False)
    OP_DESCR_QNAME: etree.QName | None = None  # to be defined in derived classes
    OP_STATE_QNAME: etree.QName | None = None # to be defined in derived classes

    def __init__(self,  # noqa: PLR0913
                 handle: str,
                 operation_target_handle: str,
                 operation_handler: ExecuteHandler,
                 timeout_handler: TimeoutHandler | None = None,
                 coded_value: CodedValue | None = None,
                 delayed_processing: bool = True,
                 log_prefix: str | None = None):
        """Construct a OperationDefinitionBase.

        :param handle: the handle of the operation itself.
        :param operation_target_handle: the handle of the modified data (MdDescription)
        :param coded_value: a pmtypes.CodedValue instance
        :param delayed_processing: if True, device returns WAIT, and sends notifications WAIT, STARTED, FINISHED/FAILED
                                   if False, device returns FINISHED/FAILED, and sends same state in notification
        """
        self._logger = loghelper.get_logger_adapter(f'sdc.device.op.{self.__class__.__name__}', log_prefix)
        self._mdib: ProviderMdib | None = None
        self._operation_entity = None
        self.handle: str = handle
        self.operation_target_handle: str = operation_target_handle
        # documentation of operation_target_handle:
        # A HANDLE reference this operation is targeted to. In case of a single state this is the HANDLE of the
        # descriptor.
        # In case that multiple states may belong to one descriptor (pm:AbstractMultiState),
        # OperationTarget is the HANDLE of one of the state instances (if the state is modified by the operation).
        self._operation_handler = operation_handler
        self._timeout_handler = timeout_handler
        self._coded_value = coded_value
        self.delayed_processing = delayed_processing
        self.calls = []  # record when operation was called
        self.last_called_time = None

    @property
    def descriptor_container(self) -> AbstractDescriptorProtocol:  # noqa: D102
        return self._operation_entity.descriptor

    def execute_operation(self,
                          soap_request: ReceivedSoapMessage,
                          operation_request: AbstractSet) -> ExecuteResult:
        """Execute the operation itself.

        This method calls the provided operation_handler.
        """
        self.calls.append((time.time(), soap_request))
        execute_result = self._operation_handler(ExecuteParameters(self, operation_request, soap_request))
        self.current_request = soap_request
        self.current_argument = operation_request.argument
        self.last_called_time = time.time()
        return execute_result

    def check_timeout(self):
        """Set on_timeout observable if timeout is detected."""
        if self.last_called_time is None:
            return
        if self._operation_entity.descriptor.InvocationEffectiveTimeout is None:
            return
        age = time.time() - self.last_called_time
        if age < self._operation_entity.descriptor.InvocationEffectiveTimeout:
            return
        if self._timeout_handler is not None:
            self._timeout_handler(self)
        self.on_timeout = True  # let observable fire

    def set_mdib(self, mdib: ProviderMdib, parent_descriptor_handle: str):
        """Set mdib reference.

        The operation needs to know the mdib that it operates on.
        This method  is called by SubscriptionManager on registration.
        Needs to be implemented by derived classes if specific things have to be initialized.
        """
        if self._mdib is not None:
            raise ApiUsageError('Mdib is already set')
        self._mdib = mdib
        self._logger.log_prefix = mdib.log_prefix  # use same prefix as mdib for logging
        self._operation_entity = self._mdib.entities.by_handle(self.handle)
        if self._operation_entity is not None:
            # there is already a descriptor
            self._logger.debug('descriptor for operation "%s" is already present, re-using it', self.handle)
        else:
            self._operation_entity = self._mdib.entities.new_entity(self.OP_DESCR_QNAME,
                                                                    self.handle,
                                                                    parent_descriptor_handle )
            self._init_operation_descriptor_container()
            with self._mdib.descriptor_transaction() as mgr:
                mgr.write_entity(self._operation_entity)

    def _init_operation_descriptor_container(self):
        self._operation_entity.descriptor.OperationTarget = self.operation_target_handle
        if self._coded_value is not None:
            self._operation_entity.descriptor.Type = self._coded_value

    def set_operating_mode(self, mode: OperatingMode):
        """Set OperatingMode member in state in transaction context."""
        entity = self._mdib.entities.by_handle(self.handle)
        entity.state.OperatingMode = mode

        with self._mdib.operational_state_transaction() as mgr:
            mgr.write_entity(entity)

    def __str__(self):
        code = None if self._operation_entity is None else self._operation_entity.descriptor.Type
        return (f'{self.__class__.__name__} handle={self.handle} code={code} '
               f'operation-target={self.operation_target_handle}')


class SetStringOperation(OperationDefinitionBase):
    """Implementation of SetString operation."""

    OP_DESCR_QNAME = pm.SetStringOperationDescriptor
    OP_STATE_QNAME = pm.SetStringOperationState


class SetValueOperation(OperationDefinitionBase):
    """Implementation of SetValue operation."""

    OP_DESCR_QNAME = pm.SetValueOperationDescriptor
    OP_STATE_QNAME = pm.SetValueOperationState


class SetContextStateOperation(OperationDefinitionBase):
    """Implementation of SetContextOperation."""

    OP_DESCR_QNAME = pm.SetContextStateOperationDescriptor
    OP_STATE_QNAME = pm.SetContextStateOperationState


class ActivateOperation(OperationDefinitionBase):
    """Parameters of an ActivateOperation."""

    OP_DESCR_QNAME = pm.ActivateOperationDescriptor
    OP_STATE_QNAME = pm.ActivateOperationState


class SetAlertStateOperation(OperationDefinitionBase):
    """Parameters of an SetAlertStateOperation."""

    OP_DESCR_QNAME = pm.SetAlertStateOperationDescriptor
    OP_STATE_QNAME = pm.SetAlertStateOperationState


class SetComponentStateOperation(OperationDefinitionBase):
    """Parameters of an SetComponentStateOperation."""

    OP_DESCR_QNAME = pm.SetComponentStateOperationDescriptor
    OP_STATE_QNAME = pm.SetComponentStateOperationState


class SetMetricStateOperation(OperationDefinitionBase):
    """Parameters of an SetMetricStateOperation."""

    OP_DESCR_QNAME = pm.SetMetricStateOperationDescriptor
    OP_STATE_QNAME = pm.SetMetricStateOperationState


# mapping of states: xsi:type information to classes
# find all classes in this module that have a member "OP_DESCR_QNAME"
_classes = inspect.getmembers(sys.modules[__name__],
                              lambda member: inspect.isclass(member) and member.__module__ == __name__)
_classes_with_qname = [c[1] for c in _classes if hasattr(c[1], 'OP_DESCR_QNAME') and c[1].OP_DESCR_QNAME is not None]
# make a dictionary from found classes: (Key is OP_DESCR_QNAME, value is the class itself
_operation_lookup_by_type = {c.OP_DESCR_QNAME: c for c in _classes_with_qname}


def get_operation_class(q_name: etree.QName) -> type[OperationDefinitionBase]:
    """:param q_name: a QName instance"""
    return _operation_lookup_by_type.get(q_name)



from time import perf_counter, sleep


class IntervalTimer:
    """IntervalTimer is a timer that does not drift (but is has jitter)."""

    def __init__(self, period_in_seconds: float):
        self._period = period_in_seconds
        self._next_interval_start = perf_counter() + self._period

    def set_period(self, period: float):
        """Set timer period."""
        self._period = period

    def reset(self):
        """Reset timer."""
        self._next_interval_start = perf_counter() + self._period

    def wait_next_interval_begin(self) -> float:
        """Return when enough time is elapsed.

        :return: 0.0 if timer is in scheduled plan, otherwise seconds how far timer is behind schedule.
        """
        behind_schedule = 0.0
        dt_remaining = self.remaining_time()
        if dt_remaining <= 0:
            behind_schedule = abs(dt_remaining)
        else:
            sleep(dt_remaining)
        self._next_interval_start += self._period
        return behind_schedule

    def remaining_time(self) -> float:
        """Return remaining time until next period."""
        return self._next_interval_start - perf_counter()



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, Type, TYPE_CHECKING

from .dpwshostedservice import DPWSHostedService

if TYPE_CHECKING:
    from .porttypes.porttypebase import DPWSPortTypeBase


@dataclass(frozen=True)
class HostedServices:
    """This is a container for all instantiated hosted services and port types.
    The references to the services are for convenience."""
    dpws_hosted_services: Dict[str, DPWSHostedService]
    get_service: Type[DPWSPortTypeBase]
    set_service: Type[DPWSPortTypeBase] = None
    context_service: Type[DPWSPortTypeBase] = None
    description_event_service: Type[DPWSPortTypeBase] = None
    state_event_service: Type[DPWSPortTypeBase] = None
    waveform_service: Type[DPWSPortTypeBase] = None
    containment_tree_service: Type[DPWSPortTypeBase] = None
    localization_service: Type[DPWSPortTypeBase] = None


def mk_dpws_hosts(sdc_device, components, dpws_hosted_service_cls, subscription_managers: dict) -> (dict, dict):
    dpws_services = {}
    services_by_name = {}
    for host_name, service_cls_list in components.hosted_services.items():
        services = []
        for service_cls in service_cls_list:
            service = service_cls(sdc_device)
            services.append(service)
            services_by_name[service.port_type_name.localname] = service
        subscription_manager = subscription_managers.get(host_name)
        hosted = dpws_hosted_service_cls(sdc_device, subscription_manager, host_name, services)
        dpws_services[host_name] = hosted
    return dpws_services, services_by_name


def mk_all_services(sdc_device, components, subscription_managers) -> HostedServices:
    # register all services with their endpoint references acc. to structure in components
    dpws_hosts, services_by_name = mk_dpws_hosts(sdc_device, components, DPWSHostedService, subscription_managers)
    hosted_services = HostedServices(dpws_hosts,
                                     services_by_name['GetService'],
                                     set_service=services_by_name.get('SetService'),
                                     context_service=services_by_name.get('ContextService'),
                                     description_event_service=services_by_name.get('DescriptionEventService'),
                                     state_event_service=services_by_name.get('StateEventService'),
                                     waveform_service=services_by_name.get('WaveformService'),
                                     containment_tree_service=services_by_name.get('ContainmentTreeService'),
                                     localization_service=services_by_name.get('LocalizationService')
                                     )
    return hosted_services



from sdc11073.dispatch import DispatchKey
from sdc11073.namespaces import PrefixesEnum
from sdc11073.xml_types import msg_qnames
from sdc11073.xml_types.actions import Actions

from .serviceclientbase import HostedServiceClient


class WaveformClient(HostedServiceClient):
    """Client for WaveformService."""

    port_type_name = PrefixesEnum.SDC.tag('WaveformService')
    notifications = (DispatchKey(Actions.Waveform, msg_qnames.WaveformStream),)



from __future__ import annotations

from decimal import Decimal
from typing import Protocol, Any

from . import isoduration

STRICT_VALUE_CHECK = True


class DataConverterProtocol(Protocol):
    def to_py(self, xml_value: str):
        ...

    def to_xml(self, py_value: Any) -> str:
        ...

    def check_valid(self, py_value: Any):
        ...

    def elem_to_py(self, xml_value: str) -> Any:
        ...


class NullConverter:
    @staticmethod
    def to_py(xml_value):
        return xml_value

    @staticmethod
    def to_xml(py_value):
        return py_value

    @staticmethod
    def check_valid(py_value):
        pass

    @staticmethod
    def elem_to_py(xml_value):
        return xml_value


class ClassCheckConverter(NullConverter):
    """No conversion, only type checking"""

    def __init__(self, *klass):
        self._klass = klass

    def check_valid(self, py_value):
        if STRICT_VALUE_CHECK and py_value is not None:
            for cls in self._klass:
                if isinstance(py_value, cls):
                    return
            raise ValueError(f'Value can only be {[cls.__name__ for cls in self._klass]}, got {type(py_value)}')


class EnumConverter(NullConverter):
    """
    Converts between enums and strings
    """

    def __init__(self, klass):
        self._klass = klass

    def to_py(self, xml_value):
        value = self._klass(xml_value)
        return value

    def to_xml(self, py_value):
        return py_value.value if hasattr(py_value, 'value') else py_value

    def check_valid(self, py_value) -> bool:
        if STRICT_VALUE_CHECK and py_value is not None:
            if not isinstance(py_value, self._klass):
                raise ValueError(f'Value can only be {self._klass.__name__}, got {type(py_value)}')


class StringConverter(NullConverter):
    """Convert None to empty string, everything else is unchanged."""

    @staticmethod
    def to_py(xml_value):
        return xml_value or ''

    @staticmethod
    def check_valid(py_value) -> bool:
        if STRICT_VALUE_CHECK and py_value is not None:
            if not isinstance(py_value, str):
                raise ValueError(f'Value can only be str, got {type(py_value)}')


class ListConverter(NullConverter):
    """Each element in list is checked and converted with provided element_converter."""

    def __init__(self, element_converter):
        if not hasattr(element_converter, 'check_valid'):
            raise TypeError
        self._element_converter = element_converter

    def check_valid(self, py_value) -> bool:
        if STRICT_VALUE_CHECK and py_value is not None:
            if not isinstance(py_value, list):
                raise ValueError(f'Value must be an instance of {type(list)}, got {type(py_value)}')
            for elem in py_value:
                self._element_converter.check_valid(elem)

    def to_py(self):
        raise NotImplementedError

    def to_xml(self):
        raise NotImplementedError

    def elem_to_py(self, xml_value):
        return self._element_converter.to_py(xml_value)

    def elem_to_xml(self, py_value):
        return self._element_converter.to_xml(py_value)


class TimestampConverter(NullConverter):
    """BICEPS Timestamp.

     XML representation: integer, representing timestamp in milliseconds
     Python representation: float in seconds
    """

    @classmethod
    def to_py(cls, xml_value: str) -> float | None:
        if xml_value is None:
            return None
        return int(xml_value) / 1000

    @staticmethod
    def to_xml(py_value) -> str:
        return str(int(py_value * 1000))

    @staticmethod
    def check_valid(py_value):
        if STRICT_VALUE_CHECK and py_value is not None:
            if not isinstance(py_value, (float, int, Decimal)):
                raise ValueError(f'Timestamp can only be float, integer or Decimal, got {type(py_value)}')
            if py_value < 0:
                raise ValueError(f'Timestamp can only have positive values, got {py_value}')


class DecimalConverter(NullConverter):
    USE_DECIMAL_TYPE = True

    @classmethod
    def to_py(cls, xml_value: str) -> Decimal | int | float:
        if xml_value is None:
            return None
        if cls.USE_DECIMAL_TYPE:
            return Decimal(xml_value)
        if '.' in xml_value:
            return float(xml_value)
        return int(xml_value)

    @staticmethod
    def _float_to_xml(py_value: Decimal | int | float) -> str:
        # round value to handle float inaccuracies
        if abs(py_value) >= 100:
            xml_value = f'{round(py_value, 1):.1f}'
        elif abs(py_value) >= 10:
            xml_value = f'{round(py_value, 2):.2f}'
        else:
            xml_value = f'{round(py_value, 3):.3f}'
        return xml_value

    @classmethod
    def _decimal_to_xml(cls, py_value):
        xml_value = str(py_value)
        if 'E' in xml_value or 'e' in xml_value:
            # no exp form allowed in xml
            return cls._float_to_xml(float(py_value))
        return xml_value

    @classmethod
    def to_xml(cls, py_value):
        if isinstance(py_value, float):
            xml_value = cls._float_to_xml(py_value)
        elif isinstance(py_value, Decimal):
            xml_value = cls._decimal_to_xml(py_value)
        else:
            xml_value = str(py_value)

        if '.' in xml_value:
            # Limit number of digits, because standard says:
            # All ·minimally conforming· processors ·must· support decimal numbers with a minimum of
            # 18 decimal digits (i.e., with a ·totalDigits· of 18).
            head, tail = xml_value.split('.')
            tail = tail[:18 - len(head)]
            if tail:
                xml_value = f'{head}.{tail}'
            else:
                xml_value = head
            # remove trailing zeros after decimal point
            while '.' in xml_value and xml_value[-1] in ('0', '.'):
                xml_value = xml_value[:-1]
        return xml_value

    @staticmethod
    def check_valid(py_value):
        if STRICT_VALUE_CHECK and py_value is not None:
            if not isinstance(py_value, Decimal):
                raise ValueError(f'expected a decimal, got {type(py_value)}')


class IntegerConverter(NullConverter):
    @staticmethod
    def to_py(xml_value: str) -> int:
        if xml_value is None:
            return None
        return int(xml_value)

    @staticmethod
    def to_xml(py_value: int) -> str:
        return str(py_value)

    @staticmethod
    def check_valid(py_value):
        if STRICT_VALUE_CHECK and py_value is not None:
            if not isinstance(py_value, int):
                raise ValueError(f'expected an integer, got {type(py_value)}')


class UnsignedIntConverter(IntegerConverter):
    MAX = 1 << 32

    @classmethod
    def check_valid(cls, py_value):
        if STRICT_VALUE_CHECK and py_value is not None:
            if not (isinstance(py_value, int) or py_value < 0 or py_value > cls.MAX):
                raise ValueError(f'expected an unsigned integer, got {type(py_value)} value={py_value}')


class UnsignedLongConverter(IntegerConverter):
    MAX = 1 << 64


class BooleanConverter(NullConverter):
    @staticmethod
    def to_py(xml_value: str) -> bool:
        return xml_value in ('true', '1')

    @staticmethod
    def to_xml(py_value: bool) -> str:
        if py_value:
            return 'true'
        return 'false'

    @staticmethod
    def check_valid(py_value):
        if STRICT_VALUE_CHECK and py_value is not None:
            if not isinstance(py_value, bool):
                raise ValueError(f'expected a boolean, got {type(py_value)}')


class DurationConverter(NullConverter):
    @staticmethod
    def to_py(xml_value: str)-> isoduration.ParsedDurationType | None:
        if xml_value is None:
            return None
        return isoduration.parse_duration(xml_value)

    @staticmethod
    def to_xml(py_value: isoduration.DurationType) -> str:
        return isoduration.duration_string(py_value)

    @staticmethod
    def check_valid(py_value: Any):
        if STRICT_VALUE_CHECK and py_value is not None:
            if not isinstance(py_value, (int, float, Decimal)):
                raise ValueError(f'expected a boolean, got {type(py_value)}')



import threading
import time
from collections import namedtuple
from functools import reduce

from sdc11073 import intervaltimer
from sdc11073.loghelper import get_logger_adapter

PeriodicStates = namedtuple('PeriodicStates', 'mdib_version states')


class PeriodicReportsNullHandler:
    def __init__(self):
        """Do nothing"""

    def start(self):
        """Do nothing"""

    def stop(self):
        """Do nothing"""

    def store_metric_states(self, mdib_version, state_updates):
        """Do nothing"""

    def store_alert_states(self, mdib_version, state_updates):
        """Do nothing"""

    def store_component_states(self, mdib_version, state_updates):
        """Do nothing"""

    def store_context_states(self, mdib_version, state_updates):
        """Do nothing"""

    def store_operational_states(self, mdib_version, state_updates):
        """Do nothing"""


class PeriodicReportsHandler:
    def __init__(self, mdib, hosted_services, fixed_interval=None):
        self._periodic_reports_interval = fixed_interval
        self._mdib = mdib
        self._hosted_services = hosted_services
        self._logger = get_logger_adapter('sdc.device.pReports')
        self._periodic_reports_lock = threading.Lock()
        self._periodic_reports_thread = None

        self._periodic_metric_reports = []
        self._periodic_alert_reports = []
        self._periodic_component_state_reports = []
        self._periodic_context_state_reports = []
        self._periodic_operational_state_reports = []
        self._run_periodic_reports_thread = False
        self._timer = None

    def start(self):
        self._run_periodic_reports_thread = True
        if self._periodic_reports_interval:
            # This setting activates the simple periodic send loop, retrievability settings are ignored
            self._run_periodic_reports_thread = True
            self._periodic_reports_thread = threading.Thread(target=self._simple_periodic_reports_send_loop,
                                                             name='DevPeriodicSendLoop')
            self._periodic_reports_thread.daemon = True
            self._periodic_reports_thread.start()
        elif self._mdib.retrievability_periodic:
            # Periodic retrievability is set at least once, start handler loop
            self._run_periodic_reports_thread = True
            self._periodic_reports_thread = threading.Thread(target=self._periodic_reports_send_loop,
                                                             name='DevPeriodicSendLoop')
            self._periodic_reports_thread.daemon = True
            self._periodic_reports_thread.start()

    def stop(self):
        self._run_periodic_reports_thread = False

    def store_metric_states(self, mdib_version, state_updates):
        self._logger.debug('store %d metric states', len(state_updates))
        self._store_for_periodic_report(mdib_version, state_updates, self._periodic_metric_reports)

    def store_alert_states(self, mdib_version, state_updates):
        self._logger.debug('store %d alert states', len(state_updates))
        self._store_for_periodic_report(mdib_version, state_updates, self._periodic_alert_reports)

    def store_component_states(self, mdib_version, state_updates):
        self._logger.debug('store %d component states', len(state_updates))
        self._store_for_periodic_report(mdib_version, state_updates, self._periodic_component_state_reports)

    def store_context_states(self, mdib_version, state_updates):
        self._logger.debug('store %d context states', len(state_updates))
        self._store_for_periodic_report(mdib_version, state_updates, self._periodic_context_state_reports)

    def store_operational_states(self, mdib_version, state_updates):
        self._logger.debug('store %d operational states', len(state_updates))
        self._store_for_periodic_report(mdib_version, state_updates, self._periodic_operational_state_reports)

    def _store_for_periodic_report(self, mdib_version, state_updates, destination_list):
        copied_updates = [s.mk_copy() for s in state_updates]
        with self._periodic_reports_lock:
            destination_list.append(PeriodicStates(mdib_version, copied_updates))

    def _simple_periodic_reports_send_loop(self):
        """This is a very basic implementation of periodic reports, it only supports fixed interval.
        It does not care about retrievability settings in the mdib.
        """
        self._logger.debug('_simple_periodic_reports_send_loop start')
        time.sleep(0.1)  # start delayed
        timer = intervaltimer.IntervalTimer(period_in_seconds=self._periodic_reports_interval)
        while self._run_periodic_reports_thread:
            timer.wait_next_interval_begin()
            self._logger.debug('_simple_periodic_reports_send_loop')
            ses = self._hosted_services.state_event_service
            cs = self._hosted_services.context_service
            for reports_list, send_func, msg in \
                    [(self._periodic_metric_reports, ses.send_periodic_metric_report, 'metric'),
                     (self._periodic_alert_reports, ses.send_periodic_alert_report, 'alert'),
                     (self._periodic_component_state_reports, ses.send_periodic_component_state_report, 'component'),
                     (self._periodic_context_state_reports, cs.send_periodic_context_report, 'context'),
                     (self._periodic_operational_state_reports, ses.send_periodic_operational_state_report,
                      'operational'),
                     ]:
                tmp = None
                with self._periodic_reports_lock:
                    if reports_list:
                        tmp = reports_list[:]
                        del reports_list[:]
                if tmp:
                    self._logger.debug('send periodic %s report', msg)
                    send_func(tmp, self._mdib.mdib_version_group)

    def _periodic_reports_send_loop(self):
        """This implementation of periodic reports send loop considers retrievability settings in the mdib.
        """

        # helper for reduce
        def _next(x, y):  # pylint: disable=invalid-name
            return x if x[1].remaining_time() < y[1].remaining_time() else y

        self._logger.debug('_periodic_reports_send_loop start')
        time.sleep(0.1)  # start delayed
        # create an interval timer for each period
        timers = {}
        for period_ms in self._mdib.retrievability_periodic:
            timers[period_ms] = intervaltimer.IntervalTimer(period_in_seconds=period_ms / 1000)
        while self._run_periodic_reports_thread:
            # find timer with the shortest remaining time
            period_ms, timer = reduce(lambda x, y: _next(x, y), timers.items())  # pylint: disable=invalid-name
            timer.wait_next_interval_begin()
            self._logger.debug('_periodic_reports_send_loop {} msec timer', period_ms)
            all_handles = self._mdib.retrievability_periodic.get(period_ms, [])
            # separate them by notification types
            metrics = []
            components = []
            alerts = []
            operationals = []
            contexts = []
            for handle in all_handles:
                descr = self._mdib.descriptions.handle.get_one(handle)
                if descr.is_metric_descriptor and not descr.is_realtime_sample_array_metric_descriptor:
                    metrics.append(handle)
                elif descr.is_system_context_descriptor or descr.is_component_descriptor:
                    components.append(handle)
                elif descr.is_alert_descriptor:
                    alerts.append(handle)
                elif descr.is_operational_descriptor:
                    operationals.append(handle)
                elif descr.is_context_descriptor:
                    contexts.append(handle)

            with self._mdib.mdib_lock:
                mdib_version = self._mdib.mdib_version
                metric_states = [self._mdib.states.descriptor_handle.get_one(h).mk_copy() for h in metrics]
                component_states = [self._mdib.states.descriptor_handle.get_one(h).mk_copy() for h in components]
                alert_states = [self._mdib.states.descriptor_handle.get_one(h).mk_copy() for h in alerts]
                operational_states = [self._mdib.states.descriptor_handle.get_one(h).mk_copy() for h in operationals]
                context_states = []
                for context in contexts:
                    print(
                        f'context.Handle {context} = {len(self._mdib.context_states.descriptor_handle.get(context, []))} states')
                    context_states.extend(
                        [st.mk_copy() for st in self._mdib.context_states.descriptor_handle.get(context, [])])
            self._logger.debug('   _periodic_reports_send_loop {} metric_states', len(metric_states))
            self._logger.debug('   _periodic_reports_send_loop {} component_states', len(component_states))
            self._logger.debug('   _periodic_reports_send_loop {} alert_states', len(alert_states))
            self._logger.debug('   _periodic_reports_send_loop {} alert_states', len(alert_states))
            self._logger.debug('   _periodic_reports_send_loop {} context_states', len(context_states))
            srv = self._hosted_services.state_event_service
            if metric_states:
                periodic_states = PeriodicStates(mdib_version, metric_states)
                srv.send_periodic_metric_report(
                    [periodic_states], self._mdib.mdib_version_group)
            if component_states:
                periodic_states = PeriodicStates(mdib_version, component_states)
                srv.send_periodic_component_state_report(
                    [periodic_states], self._mdib.mdib_version_group)
            if alert_states:
                periodic_states = PeriodicStates(mdib_version, alert_states)
                srv.send_periodic_alert_report(
                    [periodic_states], self._mdib.mdib_version_group)
            if operational_states:
                periodic_states = PeriodicStates(mdib_version, operational_states)
                srv.send_periodic_operational_state_report(
                    [periodic_states], self._mdib.mdib_version_group)
            if context_states:
                ctx_srv = self._hosted_services.context_service
                periodic_states = PeriodicStates(mdib_version, context_states)
                ctx_srv.send_periodic_context_report(
                    [periodic_states], self._mdib.mdib_version_group)



"""Communication logger."""

from __future__ import annotations

import contextlib
import functools
import logging
import pathlib
import threading
import time
import warnings
from typing import Any, Callable

warnings.warn('commlog module is deprecated and will be removed in a future version.', DeprecationWarning, stacklevel=2)

DISCOVERY_IN = 'sdc_comm.discovery.in'
DISCOVERY_OUT = 'sdc_comm.discovery.out'
SOAP_REQUEST_IN = 'sdc_comm.soap.request.in'
SOAP_REQUEST_OUT = 'sdc_comm.soap.request.out'
SOAP_RESPONSE_IN = 'sdc_comm.soap.response.in'
SOAP_RESPONSE_OUT = 'sdc_comm.soap.response.out'
SOAP_SUBSCRIPTION_IN = 'sdc_comm.soap.subscription.in'
WSDL = 'sdc_comm.wsdl'

LOGGER_NAMES = (
    DISCOVERY_IN,
    DISCOVERY_OUT,
    SOAP_REQUEST_IN,
    SOAP_REQUEST_OUT,
    SOAP_RESPONSE_IN,
    SOAP_RESPONSE_OUT,
    SOAP_SUBSCRIPTION_IN,
    WSDL,
)


class IpFilter(logging.Filter):
    """Filter all messages for the specific ip address."""

    def __init__(self, ip: str):
        super().__init__()
        self.ip = ip

    def filter(self, record: logging.LogRecord) -> bool:  # noqa: D102
        return self.ip == getattr(record, 'ip_address', None)


class CommLogger:
    """Base class to make configuring comm logger easy."""

    def __init__(self):
        self.handlers: dict[str, logging.Handler] = {}

    def start(self) -> None:
        """Start logger."""
        for name, handler in self.handlers.items():
            logger = logging.getLogger(name)
            logger.addHandler(handler)
            logger.setLevel(logging.DEBUG)

    def stop(self) -> None:
        """Stop logger."""
        for name, handler in self.handlers.items():
            logger = logging.getLogger(name)
            # Ensure handlers are fully detached and closed (important on Windows)
            logger.removeHandler(handler)
            with contextlib.suppress(Exception):
                handler.close()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *args, **kwargs):  # noqa: ANN002, ANN003
        self.stop()


class DirectoryLogger(CommLogger):
    """Logger writing communication logs into a directory. Each message will be contained in a single file."""

    D_IN = 'in'
    D_OUT = 'out'

    T_UDP = 'udp'
    T_UDP_MULTICAST = 'udpB'
    T_HTTP = 'http_subscr'
    T_WSDL = 'wsdl'

    T_HTTP_REQ = 'http_req'
    T_HTTP_RESP = 'http_resp'

    def __init__(
        self,
        log_folder: str | pathlib.Path,
        log_out: bool = False,
        log_in: bool = False,
        broadcast_ip_filter: str | None = None,
    ):
        super().__init__()
        self._log_folder = pathlib.Path(log_folder)
        self._io_lock = threading.Lock()
        self._stop_called = threading.Event()

        if log_in:
            self.handlers.update(
                {
                    DISCOVERY_IN: self._GenericHandler(functools.partial(self._write_log, self.T_UDP, self.D_IN)),
                    SOAP_REQUEST_IN: self._GenericHandler(
                        functools.partial(self._write_log, self.T_HTTP_REQ, self.D_IN),
                    ),
                    SOAP_RESPONSE_IN: self._GenericHandler(
                        functools.partial(self._write_log, self.T_HTTP_RESP, self.D_IN),
                    ),
                    SOAP_SUBSCRIPTION_IN: self._GenericHandler(
                        functools.partial(self._write_log, self.T_HTTP, self.D_IN),
                    ),
                    WSDL: self._GenericHandler(functools.partial(self._write_log, self.T_WSDL, self.D_IN)),
                },
            )
        if log_out:
            self.handlers.update(
                {
                    DISCOVERY_OUT: self._GenericHandler(functools.partial(self._write_log, self.T_UDP, self.D_OUT)),
                    SOAP_REQUEST_OUT: self._GenericHandler(
                        functools.partial(self._write_log, self.T_HTTP_REQ, self.D_OUT),
                    ),
                    SOAP_RESPONSE_OUT: self._GenericHandler(
                        functools.partial(self._write_log, self.T_HTTP_RESP, self.D_OUT),
                    ),
                },
            )
        if broadcast_ip_filter:
            broadcast_filter = IpFilter(broadcast_ip_filter)
            if DISCOVERY_IN in self.handlers:
                self.handlers[DISCOVERY_IN].addFilter(broadcast_filter)
            if DISCOVERY_OUT in self.handlers:
                self.handlers[DISCOVERY_OUT].addFilter(broadcast_filter)

    def start(self) -> None:
        """Start logger and create log folder if it does not exist."""
        self._log_folder.mkdir(parents=True, exist_ok=True)
        super().start()

    def stop(self) -> None:
        """Stop logger and wait for any in-flight writes."""
        with self._io_lock:
            self._stop_called.set()
        super().stop()

    def _mk_filename(self, ip_type: str, direction: str, *infos: str) -> str:
        """Create file name.

        :param ip_type: "tcp" or "udp"
        :param direction: "in" or "out"
        :param info: becomes part of filename
        :return:
        """
        assert ip_type in (
            self.T_UDP,
            self.T_UDP_MULTICAST,
            self.T_HTTP,
            self.T_HTTP_REQ,
            self.T_HTTP_RESP,
            self.T_WSDL,
        )
        assert direction in (self.D_IN, self.D_OUT)
        extension = 'wsdl' if ip_type == self.T_WSDL else 'xml'
        time_string = f'{time.time():06.3f}'[-8:]
        info_text = f'-{"-".join(infos)}' if infos else ''
        return f'{time_string}-{direction}-{ip_type}{info_text}.{extension}'

    def _write_log(self, ttype: str, direction: str, msg: bytes, *infos: str) -> None:
        with self._io_lock:
            if self._stop_called.is_set():
                return
            path = self._log_folder.joinpath(self._mk_filename(ttype, direction, *infos))
            path.write_bytes(msg)

    class _GenericHandler(logging.Handler):
        def __init__(self, emit: Callable):
            super().__init__()
            self._emit = emit

        def emit(self, record: logging.LogRecord) -> None:
            try:
                msg = self.format(record).encode()  # defaults to utf-8
                args = []
                if ip := getattr(record, 'ip_address', None):
                    args.append(ip)
                if http_method := getattr(record, 'http_method', None):
                    args.append(http_method)
                self._emit(msg, *args)
            except Exception:  # noqa: BLE001
                self.handleError(record)


class StreamLogger(CommLogger):
    """Set a stream handler for each comm logger."""

    def __init__(self, stream: Any | None = None, broadcast_ip_filter: str | None = None):
        super().__init__()
        for name in LOGGER_NAMES:
            self.handlers[name] = self._get_handler(stream)

        if broadcast_ip_filter:
            broadcast_filter = IpFilter(broadcast_ip_filter)
            if DISCOVERY_IN in self.handlers:
                self.handlers[DISCOVERY_IN].addFilter(broadcast_filter)
            if DISCOVERY_OUT in self.handlers:
                self.handlers[DISCOVERY_OUT].addFilter(broadcast_filter)

    @staticmethod
    def _get_handler(stream: Any | None) -> logging.StreamHandler:
        handler = logging.StreamHandler(stream=stream)
        handler.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(message)s'))
        return handler
